@@ -1,5 +1,6 @@
-import requests, uuid, json, urllib.parse, os
+import requests, uuid, json, urllib.parse, os, time, random
 from metacat.auth import TokenLib, TokenAuthClientMixin
+from metacat.common import HTTPClient
 
 import urllib3
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
@@ -32,6 +33,9 @@ class NotFoundError(ServerError):
     def __init__(self, url, message):
         ServerError.__init__(self, url, 404, message, "")
 
+class TimeoutError(Exception):
+    pass
+
 def to_bytes(x):
     if not isinstance(x, bytes):
         x = x.encode("utf-8")
@@ -41,73 +45,14 @@ def to_str(x):
     if isinstance(x, bytes):
         x = x.decode("utf-8")
     return x
-    
-class HTTPClient(object):
 
-    def __init__(self, server_url, token=None):
-        self.ServerURL = server_url
-        self.Token = token
-
-    def get(self, uri_suffix, none_if_not_found=False):
-        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
-        url = "%s%s" % (self.ServerURL, uri_suffix)
-        #print("url:", url)
-        headers = {}
-        if self.Token is not None:
-            headers["X-Authentication-Token"] = self.Token.encode()
-        response = requests.get(url, headers =headers)
-        #print("response status code:", response.status_code)
-        if response.status_code != 200:
-            if none_if_not_found and response.status_code == 404:
-                return None
-            elif response.status_code == 404:
-                raise NotFoundError(url, response.text)
-            else:
-                raise APIError(url, response.status_code, response.text)
-        if response.headers.get("Content-Type", "").startswith("text/json"):
-            data = json.loads(response.text)
-        else:
-            data = response.text
-        #print("   data:", data)
-        return data
-        
-    def post(self, uri_suffix, data):
-        #print("post_json: data:", type(data), data)
-        
-        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
-        
-        if data is None or isinstance(data, (dict, list)):
-            data = json.dumps(data)
-        else:
-            data = to_bytes(data)
-        #print("post_json: data:", type(data), data)
-            
-        url = "%s%s" % (self.ServerURL, uri_suffix)
-        
-        headers = {}
-        if self.Token is not None:
-            headers["X-Authentication-Token"] = self.Token.encode()
-        #print("HTTPClient.post_json: url:", url)
-        #print("HTTPClient.post_json: data:", data)
-        
-        response = requests.post(url, data = data, headers = headers)
-        if response.status_code != 200:
-            if response.status_code == 404:
-                raise NotFoundError(url, response.text)
-            else:
-                raise APIError(url, response.status_code, response.text)
-        #print("response.text:", response.text)
-        if response.headers.get("Content-Type", "").startswith("text/json"):
-            data = json.loads(response.text)
-        else:
-            data = response.text
-        return data
-        
 class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
     
     DefaultWorkerIDFile = ".data_dispatcher_worker_id"
     
-    def __init__(self, server_url=None, auth_server_url=None, worker_id=None, worker_id_file=None, token = None, token_file = None):
+    def __init__(self, server_url=None, auth_server_url=None, worker_id=None, worker_id_file=None, 
+            token = None, token_file = None, token_library = None, 
+            cpu_site="DEFAULT", timeout=300):
         
         """Initializes the DataDispatcherClient object
 
@@ -117,11 +62,13 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
             worker_id_file (str): File path to read/store the worker ID. 
                 Default: <cwd>/.data_dispatcher_worker_id
             worker_id (str): Worker ID to use when reserving next file. If unspecified, will be read from the worker ID file.
+            cpu_site (str): Name of the CPU site where the client is running, optional. Will be used when reserving project files.
+            timeout (float or int): Number of seconds to wait for a response.
         """
         
         server_url = server_url or os.environ.get("DATA_DISPATCHER_URL")
         auth_server_url = auth_server_url or os.environ.get("DATA_DISPATCHER_AUTH_URL")
-        TokenAuthClientMixin.__init__(self, server_url, token, token_file, auth_url=auth_server_url)
+        TokenAuthClientMixin.__init__(self, server_url, auth_server_url, token=token, token_file=token_file, token_library=token_library)
 
         #print("DataDispatcherClient: url:", server_url)
 
@@ -135,11 +82,17 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
             if worker_id_file:
                 open(worker_id_file, "w").write(worker_id)
         self.WorkerID = worker_id
+        self.CPUSite = cpu_site
+
+        HTTPClient.__init__(self, server_url, token=self.token(), timeout=timeout)
         
-        HTTPClient.__init__(self, server_url, self.token())
-            
-        #print("DataDispatcherClient: token:", token, "  user:", token["user"])
-        
+    @staticmethod
+    def random_worker_id(prefix=""):
+        """
+        Static method to generate random worker id
+        """
+        return prefix + uuid.uuid4().hex[:8]
+
     def gen_worker_id(self):
         u = uuid.uuid4().bytes
         n = len(u)//4
@@ -169,24 +122,40 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
         return worker_id
         
     def version(self):
+        """Returns the server version as a string
+        """
         return self.get("version")
 
     #
     # projects
     #
-    def create_project(self, files, common_attributes={}, project_attributes={}, query=None):
+    
+    DEFAULT_IDLE_TIMEOUT = 72*3600      # 72 hors
+    
+    def create_project(self, files, common_attributes={}, project_attributes={}, query=None, worker_timeout=None,
+            idle_timeout = DEFAULT_IDLE_TIMEOUT):
         """Creates new project
         
-        Args:
-            files (list): each item in the list is either a dictionary with keys: "namespace", "name", "attributes" (optional) or a string "namespace:name"
+        Parameters
+        ----------
+        files : list
+            Each item in the list is either a dictionary with keys: "namespace", "name", "attributes" (optional) or a string "namespace:name"
+        common_attributes : dict
+            attributes to attach to each file, will be overridden by the individual file attribute values with the same key
+        project_attributes : dict
+            attriutes to attach to the new project
+        query : str 
+            query used to create the file list, optional. If specified, the query string will be added to the project as the attribute
+        worker_timeout : int or float
+            If not None, all file handles will be automatically released if allocated by same worker for longer than the ``worker_timeout`` seconds
+        idle_timeout : int or float
+            If there is no file reserve/release activity for the specified time interval, the project goes into "abandoned" state.
+            Default is 72 hours (3 days). If set to None, the project remains active until complete.
 
-        Keyword Arguments:
-            common_attributes (dict): attributes to attach to each file, will be overridden by the individual file attribute values with the same key
-            project_attributes (dict): attriutes to attach to the new project
-            query (str): query used to create the file list, optional. If specified, the query string will be added to the project as the attribute.
-
-        Returns:
-            (dict) new project information
+        Returns
+        -------
+        dict
+            new project information
         """
         file_list = []
         for info in files:
@@ -204,14 +173,16 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
             file_list.append({"name":name, "namespace":namespace, "attributes":attrs})
         return self.post("create_project", json.dumps(
                 {   
-                    "files":        file_list,
+                    "files":                file_list,
                     "project_attributes":   project_attributes,
-                    "query":        query
+                    "query":                query,
+                    "worker_timeout":       worker_timeout,
+                    "idle_timeout":         idle_timeout
                 }
             )
         )
         
-    def copy_project(self, project_id, common_attributes={}, project_attributes={}):
+    def copy_project(self, project_id, common_attributes={}, project_attributes={}, worker_timeout=None):
         """Creates new project
         
         Args:
@@ -220,6 +191,7 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
         Keyword Arguments:
             common_attributes (dict): file attributes to override
             project_attributes (dict): project attributes to override
+            worker_timeout (int or float): worker timeout to override
 
         Returns:
             (dict) new project information
@@ -228,12 +200,12 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
                 {   
                     "project_id":           project_id,
                     "file_attributes":      common_attributes,
-                    "project_attributes":   project_attributes
+                    "project_attributes":   project_attributes,
+                    "worker_timeout":       worker_timeout
                 }
             )
         )
-    
-        
+
     def restart_handles(self, project_id, done=False, failed=False, reserved=False, all=False, handles=[]):
         """Restart processing of project file handles
         
@@ -277,6 +249,12 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
         """
         return self.get(f"cancel_project?project_id={project_id}")
         
+    def activate_project(self, project_id):
+        """
+        Resets the state of an abandoned project back to "active"
+        """
+        return self.get(f"activate_project?project_id={project_id}")
+
     def get_project(self, project_id, with_files=True, with_replicas=False):
         """Gets information about the project
         
@@ -315,7 +293,7 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
         else:
             return None
     
-    def list_projects(self, owner=None, state=None, not_state=None, attributes=None, with_files=True, with_replicas=False):
+    def list_projects(self, owner=None, state="active", not_state=None, attributes=None, with_files=True, with_replicas=False):
         """Lists existing projects
         
         Keyword Arguments:
@@ -341,25 +319,131 @@ class DataDispatcherClient(HTTPClient, TokenAuthClientMixin):
         args = "?" + "&".join(args) if args else ""
         return self.get(f"projects{args}")
 
-    def next_file(self, project_id, cpu_site=None, worker_id=None):
-        """Reserves next available file from the project
+    def search_projects(self, search_query, owner=None, state="active", with_files=True, with_replicas=False):
+        """Lists existing projects
         
-        Args:
-            project_id (int): project id to reserve a file from
-            cpu_site (str): optional, if specified, the file will be reserved according to the CPU/RSE proximity map
+        Arguments:
+            search_query (str): project search query in subset of MQL
         
+        Keyword Arguments:
+            owner (str): Include only projects owned by the specified user. Default: all users
+            with_files (boolean): Include information about files. Default: True
+            with_replicas (boolean): Include information about file replics. Default: False
+    
         Returns:
-            dictionary with file information, or None if no file was available to be reserved. The method does not block and always returns immediately.
-            Use `get_project()` to see if the project is done.
+            list of dictionaries with information about projects found
         """
+        
+        info = {
+            "query":    search_query,
+            "with_handles": with_files,
+            "with_replicas": with_replicas,
+        }
+        if state != "all":  info["state"] = state
+        if owner:   info["owner"] = owner
+        return self.post("search_projects", json.dumps(info))
 
-        worker_id = worker_id or self.WorkerID
+    def __next_file(self, project_id, cpu_site, worker_id):
         if worker_id is None:
             raise ValueError("DataDispatcherClient must be initialized with Worker ID")
         url_tail = f"next_file?project_id={project_id}&worker_id={worker_id}"
         if cpu_site:
             url_tail += f"&cpu_site={cpu_site}"
         return self.get(url_tail)
+
+    def next_file(self, project_id, cpu_site=None, worker_id=None, timeout=None, stagger=10):
+        """Reserves next available file from the project
+
+        Args:
+            project_id (int): project id to reserve a file from
+            cpu_site (str): optional, if specified, the file will be reserved according to the CPU/RSE proximity map
+            timeout (int or float): optional, if specified, time to wait for a file to become available. Otherwise, will wait indefinitely
+            stagger (int or float): optional, introduce a random delay between 0 and <stagger> seconds before sending first request. 
+                This will help mitigate the effect of synchronous stard of multiple workers. Default: 10
+
+        Returns:
+            Dictionary or boolean.
+            If dictionary, the dictionary contains the reserved file information.
+            If ``True``: the request timed out, but can be retried.
+            If ``False``: the project has ended.
+        """
+        worker_id = worker_id or self.WorkerID
+        cpu_site = cpu_site or self.CPUSite
+        t1 = None if timeout is None else time.time() + timeout
+        if stagger:
+            time.sleep(stagger * random.random())
+        retry = True
+        while retry:
+            reply = self.__next_file(project_id, cpu_site, worker_id)
+            info = reply.get("handle")
+            if info:
+                return info         # allocated
+            reason = reply.get("reason")
+            retry = reply["retry"]
+            if retry:
+                if t1 is None or time.time() < t1:
+                    dt = 60
+                    if t1 is not None:
+                        dt = min(dt, t1-time.time())
+                    dt0 = min(dt, 1.0)
+                    if dt > 0:
+                        time.sleep(dt0 + (dt-dt0)*random.random())
+                else:
+                    break
+        return retry            # True=try again later, False=project ended
+
+    def old_next_file(self, project_id, cpu_site=None, worker_id=None, timeout=None):
+        """Reserves next available file from the project
+        
+        Args:
+            project_id (int): project id to reserve a file from
+            cpu_site (str): optional, if specified, the file will be reserved according to the CPU/RSE proximity map
+            timeout (int or float): optional, if specified, time to wait for a file to become available. Otherwise, will wait indefinitely
+        
+        Returns:
+            Dictionary or boolean.
+            If dictionary, the dictionary contains the reserved file information.
+            If ``True``: the request timed out, but can be retried.
+            If ``False``: the project has ended.
+        """
+        worker_id = worker_id or self.WorkerID
+        cpu_site = cpu_site or self.CPUSite
+        t1 = None if timeout is None else time.time() + timeout
+        retry = True
+        if stagger:
+            time.sleep(random.random() * stagger)
+        while retry:
+            reply = self.__next_file(project_id, cpu_site, worker_id)
+            info = reply.get("handle")
+            if info:
+                return info         # allocated
+            retry = reply["retry"]
+            if retry:
+                if t1 is None or time.time() < t1:
+                    dt = 60
+                    if t1 is not None:
+                        dt = min(dt, t1-time.time())
+                    if dt > 0:
+                        time.sleep(max(1.0, random.random() * dt))
+                else:
+                    break
+        return retry
+
+    def reserved_files(self, project_id, worker_id=None):
+        """Returns list of file handles reserved in the project by given worker
+        
+        Args:
+            project_id (int): Project id
+            worker_id (str or None): Worker id. If None, client's worker id will be used
+    
+        Returns:
+            list of dictionaries with the file handle information
+        """
+        worker_id = worker_id or self.WorkerID
+        project = self.get_project(project_id)
+        return [h for h in project["file_handles"]
+            if h["state"] == "reserved" and h["worker_id"] == worker_id
+        ]
         
     def get_file(self, namespace, name):
         """Gets information about a file
