@@ -1,19 +1,22 @@
-import getopt, json, time, pprint, textwrap
+import getopt, json, time, pprint, textwrap, sys
 from datetime import datetime
 from metacat.webapi import MetaCatClient
-from metacat.db import DBFile
 from .ui_lib import pretty_json, parse_attrs, print_handles
 
 from .cli import CLI, CLICommand, InvalidOptions, InvalidArguments
 
 class CreateCommand(CLICommand):
     
-    Opts = "q:c:l:j:A:a:"
+    Opts = "q:c:l:j:A:a:t:p:w:"
     Usage = """[options] [inline MQL query]         -- create project
 
         Use an inline query or one of the following to provide file list:
         -l (-|<flat file with file list>)               - read "namespace:name [attr=value ...]" lines from the file 
         -j (-|<JSON file with file list>)               - read JSON file with file list {"namespace":...,"name"..., "attributes":{...}}
+
+        -w (<worker timeout>[s|m|h|d] | none)           - worker timeout in seconds (minutes, hours, days), default: 12 hours
+        -t (<idle timeout>[s|m|h|d] | none)             - worker timeout in seconds (minutes, hours, days), default: 72 hours
+
         -q <file with MQL query>                        - read MQL query from file instead
         -c <name>[,<name>...]                           - copy metadata attributes from the query results, 
                                                           use only with -q or inline query. Otherwise ignored
@@ -27,6 +30,9 @@ class CreateCommand(CLICommand):
                                                           pprint or just project id (default)
     """
     
+    FileProperties = "fid,namespace,name,checksums,size,creator,created_timestamp,parents,children,datasets".split(',') # from metacat.dbobjects2
+
+    
     def __call__(self, command, client, opts, args):
         if sum([len(args) > 0, "-l" in opts, "-j" in opts, "-q" in opts]) != 1:
             raise InvalidOptions("Use either -q or -l or -j to provide the file list")
@@ -55,6 +61,7 @@ class CreateCommand(CLICommand):
                 
         if query:
             files = MetaCatClient().query(query, with_metadata = "-c" in opts)
+            files = list(files)
             for info in files:
                 info["attributes"] = common_attrs.copy()
             
@@ -68,7 +75,7 @@ class CreateCommand(CLICommand):
                     #print(info["metadata"])
                     attrs = {}
                     for k in fields:
-                        if '.' not in k and k in DBFile.Properties:
+                        if '.' not in k and k in self.FileProperties:
                             attrs[k] = info.get(k)
                         else:
                             attrs[k] = info["metadata"].get(k)
@@ -95,10 +102,35 @@ class CreateCommand(CLICommand):
                     did, rest = (tuple(line.split(None, 1)) + ("",))[:2]
                     namespace, name = did.split(":", 1)
                     files.append({"namespace":namespace, "name":name, "attributes":parse_attrs(rest)})
+                    
+        worker_timeout = opts.get("-w")
+        if worker_timeout is not None:
+            mult = 1
+            if worker_timeout[-1].lower() in "smhd":
+                worker_timeout, unit = worker_timeout[:-1], worker_timeout[-1].lower()
+                mult = {'s':1, 'm':60, 'h':3600, 'd':24*3600}[unit]
+            worker_timeout = float(worker_timeout)*mult
+        else:
+            worker_timeout = 12*3600
+
+        idle_timeout = opts.get("-t")
+        if idle_timeout is not None:
+            mult = 1
+            if idle_timeout[-1].lower() in "smhd":
+                idle_timeout, unit = idle_timeout[:-1], idle_timeout[-1].lower()
+                mult = {'s':1, 'm':60, 'h':3600, 'd':24*3600}[unit]
+            idle_timeout = float(idle_timeout)*mult
+        else:
+            idle_timeout = 72*3600
 
         #print("files:", files)
         #print("calling API.client.create_project...")
-        info = client.create_project(files, common_attributes=common_attrs, project_attributes=project_attrs, query=query)
+        #print("")
+        if not files:
+            print("Empty file list", file=sys.stderr)
+            sys.exit(1)
+        info = client.create_project(files, common_attributes=common_attrs, project_attributes=project_attrs, query=query, 
+                    worker_timeout=worker_timeout, idle_timeout=idle_timeout)
         printout = opts.get("-p", "id")
         if printout == "json":
             print(pretty_json(info))
@@ -109,7 +141,7 @@ class CreateCommand(CLICommand):
 
 class CopyCommand(CLICommand):
     MinArgs = 1
-    Opts = "A:a:"
+    Opts = "A:a:t:p:"
     Usage = """[options] <project id>               -- copy project
 
         -A @<file.json>                                 - JSON file with project attributes to override
@@ -117,6 +149,8 @@ class CopyCommand(CLICommand):
         -a @<file.json>                                 - JSON file with file attributes to override
         -a "name=value name=value ..."                  - file attributes to override
 
+        -t <worker timeout>|none                        - worker timeout to override
+
         -p (json|pprint|id)                             - print created project info as JSON, 
                                                           pprint or just project id (default)
     """
@@ -140,8 +174,14 @@ class CopyCommand(CLICommand):
             else:
                 project_attrs = parse_attrs(attr_src)
 
+        worker_timeout = opts.get("-t")
+        if worker_timeout == "none":    
+            worker_timeout = None
+        if worker_timeout is not None:
+            worker_timeout = float(worker_timeout)
+
         #print("files:", files)
-        info = client.copy_project(project_id, common_attributes=common_attrs, project_attributes=project_attrs)
+        info = client.copy_project(project_id, common_attributes=common_attrs, project_attributes=project_attrs, worker_timeout=worker_timeout)
         printout = opts.get("-p", "id")
         if printout == "json":
             print(pretty_json(info))
@@ -155,19 +195,22 @@ class RestartCommand(CLICommand):
 
     MinArgs = 1
     Opts = "adfr"
-    Usage = """[options] <project_id> [<did> ...]   -- restart project handles
-    restart [options] <project_id>                  -- restart project files by handle state
-        -f                                              - restart failed files
-        -d                                              - restart done files
-        -r                                              - unreserve reserved files
-        -a                                              - restart all files (same as -f -d -r)
-        
-    restart <project_id> <namespace>:<name> ...     -- restart specific files, options above are ignored
+    Usage = """                                     -- restart project handles
+    restart <project_id> <DID> [...]                  -- restart specific handles by DIDs
+
+    restart [selection] <project_id>                  -- restart handles by state:
+        -f                                              - restart failed handles
+        -d                                              - restart done handles
+        -r                                              - unreserve reserved handles
+        -a                                              - same as -f -d -r
     """
 
     def __call__(self, command, client, opts, args):
         project_id = args[0]
         if args[1:]:
+            for did in args[1:]:
+                if len(did.split(':', 1)) != 2:
+                    raise InvalidArguments("Invalid DID format: %s" % (did,))
             client.restart_handles(project_id, handles=args[1:])
         else:
             handle_states = []
@@ -183,7 +226,19 @@ class RestartCommand(CLICommand):
             
             handle_states = {s:True for s in handle_states}
             client.restart_handles(project_id, **handle_states)
+            
+class ActivateCommand(CLICommand):
 
+    MinArgs = 1
+    Opts = "j"
+    Usage = """[-j] <project_id>                    -- re-activate an abandoned project
+        -j                                              - print project info as JSON
+    """
+
+    def __call__(self, command, client, opts, args):
+        project_id = int(args[0])
+        client.activate_project(project_id)
+            
 class ShowCommand(CLICommand):
     
     Opts = "arjf:"
@@ -205,6 +260,9 @@ class ShowCommand(CLICommand):
     def __call__(self, command, client, opts, args):
         project_id = args[0]
         info = client.get_project(project_id, with_files=True, with_replicas=True)
+        if info is None:
+            print("Project", project_id, "not found")
+            sys.exit(1)
         if "-a" in opts:
             attrs = info.get("attributes", {})
             if "-j" in opts:
@@ -238,6 +296,8 @@ class ShowCommand(CLICommand):
                 print("Ended:              ", ended_timestamp)
                 print("Query:              ", textwrap.indent(info.get("query") or "", " "*10).lstrip())
                 print("Status:             ", info["state"])
+                print("Worker timeout:     ", info.get("worker_timeout"))
+                print("Idle timeout:       ", info.get("idle_timeout"))
                 print("Project Attributes: ")
                 for k, v in info.get("attributes", {}).items():
                     if not isinstance(v, (int, float, str, bool)):
@@ -253,12 +313,12 @@ class ListCommand(CLICommand):
     Usage = """[options]                            -- list projects
             -j                                          - JSON output
             -u <owner>                                  - filter by owner
-            -s <state>                                  - filter by state
+            -s <state>                                  - filter by state, default: active projects only
             -a "name=value name=value ..."              - filter by attributes
     """
 
     def __call__(self, command, client, opts, args):
-        state = opts.get("-s")
+        state = opts.get("-s", "active")
         attributes = None
         if "-a" in opts:
             attributes = parse_attrs(opts["-a"])
@@ -286,6 +346,51 @@ class ListCommand(CLICommand):
                 print("%-15s %-15s %19s %15s %17s" % (prj["project_id"], prj["owner"], ct, prj["state"], counts))
             print("%s %s %s %s %s" % ("-"*15, "-"*15, "-"*19, "-"*15, "-"*17))
 
+class SearchCommand(CLICommand):
+    Opts = "ju:s:"
+    Usage = """[options] (-q (<query file>|-) |<search query>)            -- search projects
+            -j                                          - JSON output
+            -u <owner>                                  - filter by owner
+            -s (<state>|all)                            - filter by state, default: active projects only
+            -q <query file>|-                           - read the search query from the file or stdin if "-"
+    """
+
+    def __call__(self, command, client, opts, args):
+        state = opts.get("-s", "active")
+        owner = opts.get("-u")
+        if "-q" in opts:
+            query_file = opts["-q"]
+            if query_file == "-":
+                query = sys.stdin.read()
+            else:
+                query = open(query_file, "r").read()
+        else:
+            if not args:
+                raise InvalidArguments("Search query is not specified")
+            query = " ".join(args)
+        lst = client.search_projects(query, state=state, owner=owner, with_files=True, with_replicas=False)
+        if "-j" in opts:
+            print(pretty_json(list(lst)))
+        else:
+            print("%-15s %-15s %-19s %-15s %17s" % ("project id", "owner", "created", "state", "done/failed/files"))
+            print("%s %s %s %s %s" % ("-"*15, "-"*15, "-"*19, "-"*15, "-"*17))
+            for prj in lst:
+                ct = time.localtime(prj["created_timestamp"])
+                ct = time.strftime("%Y-%m-%d %H:%M:%S", ct)
+            
+                nfiles = len(prj["file_handles"])
+                ready_files = reserved_files = failed_files = done_files = available_files = 0
+                for h in prj["file_handles"]:
+                    if h["state"] == "ready":
+                        ready_files += 1
+                    elif h["state"] == "done":
+                        done_files += 1
+                    elif h["state"] == "failed":
+                        failed_files += 1
+                counts = "%d/%d/%d" % (done_files, failed_files, nfiles)
+                print("%-15s %-15s %19s %15s %17s" % (prj["project_id"], prj["owner"], ct, prj["state"], counts))
+            print("%s %s %s %s %s" % ("-"*15, "-"*15, "-"*19, "-"*15, "-"*17))
+
 class CancelCommand(CLICommand):
     
     Opts = "j"
@@ -312,8 +417,10 @@ ProjectCLI = CLI(
     "copy",     CopyCommand(),
     "show",     ShowCommand(),
     "list",     ListCommand(),
+    "search",   SearchCommand(),
+    "restart",  RestartCommand(),
+    "activate", ActivateCommand(),
     "cancel",   CancelCommand(),
-    "delete",   DeleteCommand(),
-    "restart",  RestartCommand()
+    "delete",   DeleteCommand()
 )
 
